/* core.c
 * This file is part of Vara digital painting program
 * Copyright 2021, 2023 Nandakumar Edamana
 * Distributed under GNU General Public License, version 3
 */ 

#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>

#include "core.h"

#include <math.h>
#include "pixel.h"
struct Pixel PIXEL_BLACK = (struct Pixel) { {0, 0, 0, 1} };
struct Pixel PIXEL_GRAY = (struct Pixel) { {0.5, 0.5, 0.5, 1} };
struct Pixel PIXEL_RED = (struct Pixel) { {1, 0, 0, 1} };
struct Pixel PIXEL_VPBG = (struct Pixel) { {0.2, 0.2, 0.2, 1} };
struct Pixel PIXEL_WHITE = (struct Pixel) { {1, 1, 1, 1} };
struct Pixel PIXEL_TRANSP = (struct Pixel) { {0, 0, 0, 0} };
struct Pixel PIXEL_CHECKER_DARK = (struct Pixel) { {0.5, 0.5, 0.5, 1} };
struct Pixel PIXEL_CHECKER_LIGHT = (struct Pixel) { {0.8, 0.8, 0.8, 1} };
double AA_RADIUS_CIRCLE2 = sqrt(2);
double AA_MAXDIFF_LINE = (double) 0.8;
Rectangle rectangle_new_from_boundaries(int x1, int y1, int x2, int y2)
{
	return region_new(x1, y1, (x2 - x1) + 1, (y2 - y1) + 1);
}

_Bool region_is_zero(Rectangle region)
{
	return (region.w <= 0) || (region.h <= 0);
}

Rectangle region_new(int x, int y, int w, int h)
{
	assert(w >= 0); /* core.ngg:97 */
	assert(h >= 0); /* core.ngg:98 */
	Rectangle region = rectangle_default();
	region.x = x;
	region.y = y;
	region.w = w;
	region.h = h;

	return region;
}

void image_construct(Image *this, int width, int height, int depth)
{
	assert(width > 0); /* core.ngg:126 */
	assert(height > 0); /* core.ngg:127 */
	assert(depth > 0); /* core.ngg:128 */
	this->width = width;
	this->height = height;
	this->depth = depth;
	double *_ngg_tmp_0 = malloc((width * height) * PIXSTRIDE * sizeof(double));
	if(_ngg_tmp_0 == NULL) { perror(NULL); exit(EXIT_FAILURE); }

	this->buf = _ngg_tmp_0;
}

Image* image_Clone(Image *this)
{

	Image * newimg = malloc(sizeof(Image));
	if(newimg == NULL) { perror(NULL); exit(EXIT_FAILURE); }
	image_construct(newimg, this->width, this->height, this->depth);
	image_copy_image(newimg, this);

	return newimg;
}

void image_assert_region(Image *this, Rectangle region)
{
	assert((region.x + region.w) <= this->width); /* core.ngg:148 */
	assert((region.y + region.h) <= this->height); /* core.ngg:149 */
}

void image_copy_image(Image *this, Image *src)
{
	assert(src->width == this->width); /* core.ngg:153 */
	assert(src->height == this->height); /* core.ngg:154 */
	memcpy(this->buf, src->buf, ((this->width * this->height) * PIXSTRIDE) * (sizeof(double)));
}

Image* image_make_clone(Image *this)
{

	Image * cln = malloc(sizeof(Image));
	if(cln == NULL) { perror(NULL); exit(EXIT_FAILURE); }
	image_construct(cln, this->width, this->height, this->depth);
	image_copy_image(cln, this);

	return cln;
}

Rectangle image_copy_image_region_with_destoffset(Image *this, Image *src, Rectangle region, int destx, int desty)
{
	int y;
	image_assert_region(src, region);
	int startx = imax2(0, destx + region.x);
	int starty = imax2(0, desty + region.y);
	int maxx = imin2(this->width, (destx + region.x) + region.w);
	int maxy = imin2(this->height, (desty + region.y) + region.h);
	if((maxx <= startx) || (maxy <= starty)) {
		return region_new(0, 0, 0, 0);
	}

	assert(maxx <= this->width); /* core.ngg:185 */
	assert(maxy <= this->height); /* core.ngg:186 */
	for(y = starty; y < maxy; ++y) {
		int x;
		for(x = startx; x < maxx; ++x) {
			nan_img_set_pixel(this, x, y, nan_img_get_pixel(src, x - destx, y - desty));
		}

	}


	return rectangle_new_from_boundaries(startx, starty, maxx - 1, maxy - 1);
}

Rectangle image_copy_image_region_with_scale(Image *this, Image *src, Rectangle region, int destx, int desty, double scl)
{
	int y;
	assert(scl > 0); /* core.ngg:203 */
	image_assert_region(src, region);
	int startx = imax2(0, destx + (scl * region.x));
	int starty = imax2(0, desty + (scl * region.y));
	int x2 = region.x + region.w;
	int y2 = region.y + region.h;
	int maxx = imin2(this->width, destx + (ceil(scl * x2)));
	int maxy = imin2(this->height, desty + (ceil(scl * y2)));
	if((maxx <= startx) || (maxy <= starty)) {
		return region_new(0, 0, 0, 0);
	}

	assert(maxx <= this->width); /* core.ngg:223 */
	assert(maxy <= this->height); /* core.ngg:224 */
	for(y = starty; y < maxy; ++y) {
		int x;
		for(x = startx; x < maxx; ++x) {
			nan_img_set_pixel(this, x, y, nan_img_get_pixel_for_scaling(src, (x - destx) / scl, (y - desty) / scl));
		}

	}


	return rectangle_new_from_boundaries(startx, starty, maxx - 1, maxy - 1);
}

Rectangle image_copy_image_with_destoffset(Image *this, Image *src, int destx, int desty)
{
	int y;
	int _ngg_tmp_1;
	if(destx > 0) {
		_ngg_tmp_1 = destx;
	} else {
		_ngg_tmp_1 = 0;
	}
	int startx = _ngg_tmp_1;
	int _ngg_tmp_2;
	if(desty > 0) {
		_ngg_tmp_2 = desty;
	} else {
		_ngg_tmp_2 = 0;
	}
	int starty = _ngg_tmp_2;
	int maxx = imin2(this->width, destx + src->width);
	int maxy = imin2(this->height, desty + src->height);
	if((maxx <= startx) || (maxy <= starty)) {
		return region_new(0, 0, 0, 0);
	}

	assert(maxx <= this->width); /* core.ngg:253 */
	assert(maxy <= this->height); /* core.ngg:254 */
	for(y = starty; y < maxy; ++y) {
		int x;
		for(x = startx; x < maxx; ++x) {
			nan_img_set_pixel(this, x, y, nan_img_get_pixel(src, x - destx, y - desty));
		}

	}


	return rectangle_new_from_boundaries(startx, starty, maxx - 1, maxy - 1);
}

Rectangle image_copy_image_with_scale(Image *this, Image *src, int destx, int desty, double scl)
{
	int y;
	assert(scl > 0); /* core.ngg:272 */
	int _ngg_tmp_3;
	if(destx > 0) {
		_ngg_tmp_3 = destx;
	} else {
		_ngg_tmp_3 = 0;
	}
	int startx = _ngg_tmp_3;
	int _ngg_tmp_4;
	if(desty > 0) {
		_ngg_tmp_4 = desty;
	} else {
		_ngg_tmp_4 = 0;
	}
	int starty = _ngg_tmp_4;
	int maxx = imin2(this->width, destx + (scl * src->width));
	int maxy = imin2(this->height, desty + (scl * src->height));
	if((maxx <= startx) || (maxy <= starty)) {
		return region_new(0, 0, 0, 0);
	}

	assert(maxx <= this->width); /* core.ngg:284 */
	assert(maxy <= this->height); /* core.ngg:285 */
	for(y = starty; y < maxy; ++y) {
		int x;
		for(x = startx; x < maxx; ++x) {
			nan_img_set_pixel(this, x, y, nan_img_get_pixel_for_scaling(src, (x - destx) / scl, (y - desty) / scl));
		}

	}


	return rectangle_new_from_boundaries(startx, starty, maxx - 1, maxy - 1);
}

Rectangle image_clip_region(Image *this, Rectangle region)
{
	if(region.x < 0) {
		region.w += region.x;
		region.x = 0;
	}

	if(region.y < 0) {
		region.h += region.y;
		region.y = 0;
	}

	if((region.x + region.w) > this->width) {
		region.w = this->width - region.x;
	}

	if((region.y + region.h) > this->height) {
		region.h = this->height - region.y;
	}


	return region;
}

void image_copy_image_region(Image *this, Image *src, Rectangle region)
{
	int y;
	assert(src->width == this->width); /* core.ngg:320 */
	assert(src->height == this->height); /* core.ngg:321 */
	image_assert_region(src, region);
	for(y = region.y; y < region.y + region.h; ++y) {
		int x;
		for(x = region.x; x < region.x + region.w; ++x) {
			nan_img_set_pixel(this, x, y, nan_img_get_pixel(src, x, y));
		}

	}

}

Rectangle image_dip(Image *this, double x, double y, DrawingParams dp)
{
	switch(dp.active_tool) {
	case TOOL_BRUSH:
	{
		nan_img_thick_dot(this, x, y, 1, dp);
		break;
	}
	case TOOL_PENCIL:
	{
		nan_img_aaline_dot(this, x, y, 0, dp);
		break;
	}
	}


	return image_get_dirty_region_by_stroke(this, x, y, x, y, dp);
}

void image_draw_debug_region(Image *this, Rectangle region)
{
	int x;
	int y;
	int x2 = (region.x + region.w) - 1;
	int y2 = (region.y + region.h) - 1;
	for(x = region.x; x <= x2; ++x) {
		nan_img_set_pixel(this, x, region.y, PIXEL_RED);
		nan_img_set_pixel(this, x, y2, PIXEL_RED);
	}

	for(y = region.y; y <= y2; ++y) {
		nan_img_set_pixel(this, region.x, y, PIXEL_RED);
		nan_img_set_pixel(this, x2, y, PIXEL_RED);
	}

}

Rectangle image_get_dirty_region_by_stroke(Image *this, double lbx, double lby, double ubx, double uby, DrawingParams dp)
{
	Rectangle dirtyrect = rectangle_default();
	switch(dp.active_tool) {
	case TOOL_BRUSH:
	{
		double finalrad = effective_brush_radius(dp.brushparams.siz / 2.0, dp);
		dirtyrect = rectangle_new_from_boundaries(lbx - finalrad, lby - finalrad, ceil(ubx + finalrad), ceil(uby + finalrad));
		break;
	}
	case TOOL_PENCIL:
	{
		dirtyrect = rectangle_new_from_boundaries(lbx, lby, ceil(ubx), ceil(uby));
		break;
	}
	}


	return image_clip_region(this, dirtyrect);
}

Rectangle image_get_region(Image *this)
{
	Rectangle region = rectangle_default();
	region.w = this->width;
	region.h = this->height;

	return region;
}

void image_fill(Image *this, struct Pixel pixel)
{
	int y;
	for(y = 0; y < this->height; ++y) {
		int x;
		for(x = 0; x < this->width; ++x) {
			nan_img_set_pixel(this, x, y, pixel);
		}

	}

}

void image_fill_region(Image *this, struct Pixel pixel, Rectangle region)
{
	int y;
	image_assert_region(this, region);
	int x2 = region.x + region.w;
	int y2 = region.y + region.h;
	for(y = region.y; y < y2; ++y) {
		int x;
		for(x = region.x; x < x2; ++x) {
			nan_img_set_pixel(this, x, y, pixel);
		}

	}

}

_Bool image_is_point_outside(Image *this, int x, int y)
{
	return (((x < 0) || (y < 0)) || (x >= this->width)) || (y >= this->height);
}

Rectangle image_line(Image *this, double x1, double y1, double x2, double y2, DrawingParams dp)
{
	if((x1 == x2) && (y1 == y2)) {
		return image_dip(this, x1, y1, dp);
	}

	double m = ((double) (y2 - y1)) / (x2 - x1);
	assert(!(isnan(m))); /* core.ngg:453 */
	double lbx = 0.0;
	double ubx = lbx;
	double lby = lbx;
	double uby = lbx;
	if(x1 <= x2) {
		lbx = x1;
		ubx = x2;
	} else {
		lbx = x2;
		ubx = x1;
	}

	if(y1 <= y2) {
		lby = y1;
		uby = y2;
	} else {
		lby = y2;
		uby = y1;
	}

	switch(dp.active_tool) {
	case TOOL_BRUSH:
	{
		if((ubx - lbx) > (uby - lby)) {
			nan_img_thickline_inner_by_x(this, x1, y1, x2, y2, lbx, lby, ubx, uby, m, dp);
		} else {
			nan_img_thickline_inner_by_y(this, x1, y1, x2, y2, lbx, lby, ubx, uby, m, dp);
		}

		break;
	}
	case TOOL_PENCIL:
	{
		if((ubx - lbx) > (uby - lby)) {
			nan_img_aaline_inner_by_x_dbl(this, x1, y1, x2, y2, lbx, lby, ubx, uby, m, dp);
		} else {
			nan_img_aaline_inner_by_y_dbl(this, x1, y1, x2, y2, lbx, lby, ubx, uby, m, dp);
		}

		break;
	}
	}


	return image_get_dirty_region_by_stroke(this, lbx, lby, ubx, uby, dp);
}

void image_overlay_image(Image *this, Image *img)
{
	int y;
	assert(img->width == this->width); /* core.ngg:504 */
	assert(img->height == this->height); /* core.ngg:505 */
	for(y = 0; y < this->height; ++y) {
		int x;
		for(x = 0; x < this->width; ++x) {
			nan_img_over_pixel(this, x, y, nan_img_get_pixel(img, x, y));
		}

	}

}

void image_overlay_image_region(Image *this, Image *img, Rectangle region)
{
	image_overlay_image_region_with_alpha(this, img, 1, region);
}

void image_overlay_image_region_with_alpha(Image *this, Image *img, double alpha, Rectangle region)
{
	int y;
	assert(img->width == this->width); /* core.ngg:522 */
	assert(img->height == this->height); /* core.ngg:523 */
	image_assert_region(img, region);
	int x2 = region.x + region.w;
	int y2 = region.y + region.h;
	for(y = region.y; y < y2; ++y) {
		int x;
		for(x = region.x; x < x2; ++x) {
			int i;
			struct Pixel newpix = nan_img_get_pixel(img, x, y);
			for(i = 0; i < 4; ++i) {
				newpix.rgb[i] *= alpha;
			}

			nan_img_over_pixel(this, x, y, newpix);
		}

	}

}

void image_overlay_image_with_alpha(Image *this, Image *img, double alpha)
{
	image_overlay_image_region_with_alpha(this, img, alpha, image_get_region(this));
}

void image_overlay_image_with_offset(Image *this, Image *src, int destx, int desty)
{
	int y;
	int destxmin = imax2(0, destx);
	int destymin = imax2(0, desty);
	int destxmax = imin2(destx + src->width - 1, this->width - 1);
	int destymax = imin2(desty + src->height - 1, this->height - 1);
	for(y = destymin; y <= destymax; ++y) {
		int x;
		for(x = destxmin; x <= destxmax; ++x) {
			nan_img_over_pixel(this, x, y, nan_img_get_pixel(src, x - destx, y - desty));
		}

	}

}

void image_alphen_image_with_offset(Image *this, Image *src, int destx, int desty)
{
	int y;
	int destxmin = imax2(0, destx);
	int destymin = imax2(0, desty);
	int destxmax = imin2(destx + src->width - 1, this->width - 1);
	int destymax = imin2(desty + src->height - 1, this->height - 1);
	for(y = destymin; y <= destymax; ++y) {
		int x;
		for(x = destxmin; x <= destxmax; ++x) {
			nan_img_alphen_pixel(this, x, y, (nan_img_get_pixel(src, x - destx, y - desty)).rgb[3]);
		}

	}

}

void image_destruct(Image *this)
{
	if(this->buf) {
		free(this->buf);
	}

	this->buf = NULL;
}

double effective_brush_radius(double radius, DrawingParams dp)
{
	switch(dp.brushparams.type) {
	case BRUSHTYPE_CIRCLE1:
	{
		if(!CIRCLE1_AA_IS_INSIDE) {
			radius += AA_RADIUS_CIRCLE1;
		}

		break;
	}
	case BRUSHTYPE_CIRCLE2:
	{
		radius = AA_RADIUS_CIRCLE2 + radius;
		break;
	}
	}


	return radius;
}

Image* nan_img_new(int width, int height, int depth)
{

	Image * _ngg_tmp_5 = malloc(sizeof(Image));
	if(_ngg_tmp_5 == NULL) { perror(NULL); exit(EXIT_FAILURE); }
	image_construct(_ngg_tmp_5, width, height, depth);
	return _ngg_tmp_5;
}

void nan_img_aaline_dot(Image *img, int x, int y, double diff, DrawingParams dp)
{
	if(diff <= AA_MAXDIFF_LINE) {
		double fac = 1 - (((double) diff) / AA_MAXDIFF_LINE);
		assert(fac >= 0 && fac <= 1); /* core.ngg:498 */
		nan_img_over_pixel(img, x, y, rgbpixel_with_alpha(dp.fgcolor, fac));
	}

}

double distance(double xdist, double ydist)
{
	return sqrt((xdist * xdist) + (ydist * ydist));
}

double dist2pixmid(double x1, double y1, int x2, int y2)
{
	return distance((x2 + 0.5) - x1, (y2 + 0.5) - y1);
}

int imax2(int a, int b)
{
	int _ngg_tmp_6;
	if(a > b) {
		_ngg_tmp_6 = a;
	} else {
		_ngg_tmp_6 = b;
	}
	return _ngg_tmp_6;
}

int imin2(int a, int b)
{
	int _ngg_tmp_7;
	if(a < b) {
		_ngg_tmp_7 = a;
	} else {
		_ngg_tmp_7 = b;
	}
	return _ngg_tmp_7;
}

void nan_img_aaline_inner_by_x_dbl(Image *img, double x1, double y1, double x2, double y2, double lbx, double lby, double ubx, double uby, double m, DrawingParams dp)
{
	int x;
	for(x = (int) lbx; x <= (int) (ceil(ubx)); ++x) {
		int y;
		double xdiff = 0.0;
		if(x < lbx) {
			xdiff = lbx - x;
		} else if(x > ubx) {
			xdiff = x - ubx;
		}

		assert(xdiff >= 0 && xdiff <= 1); /* core.ngg:635 */
		double yreal = (m * (x - x1)) + y1;
		for(y = (int) lby; y <= (int) (ceil(uby)); ++y) {
			double ydiff = fabs(y - yreal);
			nan_img_aaline_dot(img, x, y, distance(xdiff, ydiff), dp);
		}

	}

}

void nan_img_aaline_inner_by_y_dbl(Image *img, double x1, double y1, double x2, double y2, double lbx, double lby, double ubx, double uby, double m, DrawingParams dp)
{
	int y;
	for(y = (int) lby; y <= (int) (ceil(uby)); ++y) {
		int x;
		double ydiff = 0.0;
		if(y < lby) {
			ydiff = lby - y;
		} else if(y > uby) {
			ydiff = y - uby;
		}

		assert(ydiff >= 0 && ydiff <= 1); /* core.ngg:664 */
		double xreal = ((y - y1) / m) + x1;
		for(x = (int) lbx; x <= (int) (ceil(ubx)); ++x) {
			double xdiff = fabs(x - xreal);
			nan_img_aaline_dot(img, x, y, distance(xdiff, ydiff), dp);
		}

	}

}

void nan_img_thickline_inner_by_x(Image *img, double x1, double y1, double x2, double y2, double lbx, double lby, double ubx, double uby, double m, DrawingParams dp)
{
	int x;
	for(x = (int) (ceil(lbx)); x <= (int) (floor(ubx)); ++x) {
		double yreal = (m * (x - x1)) + y1;
		double progress = 1;
		if(dp.brushparams.pressure_size && ((ubx - lbx) != 0)) {
			progress = (((double) x) - x1) / (x2 - x1);
		}

		nan_img_thick_dot(img, x, yreal, progress, dp);
	}

}

void nan_img_thickline_inner_by_y(Image *img, double x1, double y1, double x2, double y2, double lbx, double lby, double ubx, double uby, double m, DrawingParams dp)
{
	int y;
	for(y = (int) (ceil(lby)); y <= (int) (floor(uby)); ++y) {
		double xreal = ((y - y1) / m) + x1;
		double progress = 1;
		if(dp.brushparams.pressure_size && ((uby - lby) != 0)) {
			progress = (((double) y) - y1) / (y2 - y1);
		}

		nan_img_thick_dot(img, xreal, y, progress, dp);
	}

}

void nan_img_thick_dot(Image *img, double x, double y, double progress, DrawingParams dp)
{
	double brush_size = dp.brushparams.siz;
	double opacity = dp.brushparams.opacity;
	if(dp.brushparams.pressure_size) {
		assert((progress >= 0) && (progress <= 1)); /* core.ngg:729 */
		double curpressure = dp.oldpressure + (progress * (dp.pressure - dp.oldpressure));
		assert((curpressure >= 0) && (curpressure <= 1)); /* core.ngg:734 */
		brush_size = brush_size * curpressure;
	}

	switch(dp.brushparams.type) {
	case BRUSHTYPE_CIRCLE1:
	{
		nan_img_fill_circle1(img, x, y, brush_size, opacity, false, dp);
		break;
	}
	case BRUSHTYPE_CIRCLE1NOISE:
	{
		nan_img_fill_circle1(img, x, y, brush_size, opacity, true, dp);
		break;
	}
	case BRUSHTYPE_CIRCLE2:
	{
		nan_img_fill_circle2(img, x, y, brush_size, opacity, dp);
		break;
	}
	case BRUSHTYPE_RANDSQUARE:
	{
		nan_img_fill_random_square(img, x, y, brush_size, opacity, dp);
		break;
	}
	}

}

void nan_img_fill_circle1(Image *img, double x, double y, double brush_size, double opacity, _Bool noisy, DrawingParams dp)
{
	int j;
	double radius = brush_size / 2.0;
	if(radius <= 0) {
		return ;
	}

	double maxrad = effective_brush_radius(radius, dp);
	int aaradius = imax2(AA_RADIUS_CIRCLE1, maxrad * dp.brushparams.softness);
	for(j = (int) (y - maxrad); j <= (int) (ceil(y + maxrad)); ++j) {
		int i;
		for(i = (int) (x - maxrad); i <= (int) (ceil(x + maxrad)); ++i) {
			double currad = sqrt((pow(x - i, 2)) + (pow(y - j, 2)));
			if(currad <= maxrad) {
				double alpha = 1.0;
				double raddiff = fabs(currad - maxrad);
				if(raddiff <= aaradius) {
					alpha = raddiff / aaradius;
				}

				alpha *= opacity;
				if(noisy) {
					alpha *= ((drand48()) / NOISEDIVISOR);
				}

				assert(alpha >= 0 && alpha <= 1); /* core.ngg:793 */
				if(dp.erase) {
					nan_img_alphen_pixel(img, i, j, alpha);
				} else {
					nan_img_over_pixel(img, i, j, rgbpixel_with_alpha(dp.fgcolor, alpha));
				}

			}

		}

	}

}

void nan_img_fill_circle2(Image *img, double x, double y, double brush_size, double opacity, DrawingParams dp)
{
	int j;
	if(brush_size <= 0) {
		return ;
	}

	double realrad = brush_size / 2.0;
	double maxrad = effective_brush_radius(realrad, dp);
	double maxrad_pixmid = dist2pixmid(x, y, x + maxrad, y);
	double softradstart = maxrad_pixmid * (1 - dp.brushparams.softness);
	for(j = (int) (y - maxrad); j <= (int) (ceil(y + maxrad)); ++j) {
		int i;
		int _ngg_tmp_8;
		if(j < y) {
			_ngg_tmp_8 = j;
		} else {
			_ngg_tmp_8 = j + 1;
		}
		int ybound = _ngg_tmp_8;
		for(i = (int) (x - maxrad); i <= (int) (ceil(x + maxrad)); ++i) {
			double alpha = 0.0;
			int _ngg_tmp_9;
			if(i < x) {
				_ngg_tmp_9 = i;
			} else {
				_ngg_tmp_9 = i + 1;
			}
			int xbound = _ngg_tmp_9;
			double maxradbycurpix = distance(xbound - x, ybound - y);
			if(maxradbycurpix <= realrad) {
				alpha = 1;
			} else {
				double freespace = maxradbycurpix - realrad;
				if(freespace <= AA_RADIUS_CIRCLE2) {
					alpha = 1.0 - (freespace / AA_RADIUS_CIRCLE2);
				}

			}

			if(dp.brushparams.softness > 0) {
				double currad_pixmid = dist2pixmid(x, y, i, j);
				if((currad_pixmid <= maxrad_pixmid) && (currad_pixmid >= softradstart)) {
					double fac = 1.0 - ((currad_pixmid - softradstart) / (maxrad_pixmid - softradstart));
					alpha *= fac;
				}

			}

			alpha *= opacity;
			assert(alpha >= 0 && alpha <= 1); /* core.ngg:847 */
			if(alpha > 0) {
				if(dp.erase) {
					nan_img_alphen_pixel(img, i, j, alpha);
				} else {
					nan_img_over_pixel(img, i, j, rgbpixel_with_alpha(dp.fgcolor, alpha));
				}

			}

		}

	}

}

void nan_img_fill_random_square(Image *img, double x, double y, double brush_size, double opacity, DrawingParams dp)
{
	int j;
	int radius = brush_size / 2.0;
	if(radius <= 0) {
		return ;
	}

	int w = 2 * radius;

	Image * brushimg = malloc(sizeof(Image));
	if(brushimg == NULL) { perror(NULL); exit(EXIT_FAILURE); }
	image_construct(brushimg, w, w, 255);
	image_fill(brushimg, PIXEL_TRANSP);
	for(j = 0; j < w; ++j) {
		int i;
		for(i = 0; i < w; ++i) {
			nan_img_over_pixel(brushimg, i, j, rgbpixel_with_alpha(dp.fgcolor, opacity * ((drand48()) / NOISEDIVISOR)));
		}

	}

	if(dp.erase) {
		image_alphen_image_with_offset(img, brushimg, x - radius, y - radius);
	} else {
		image_overlay_image_with_offset(img, brushimg, x - radius, y - radius);
	}

	if(brushimg) {
		image_destruct(brushimg);
		free(brushimg);
	}

	brushimg = NULL;
}

_Bool pixelsmatch(struct Pixel px1, struct Pixel px2)
{
	int i;
	for(i = 0; i < 4; ++i) {
		if(px1.rgb[i] != px2.rgb[i]) {
			return false;
		}

	}


	return true;
}

void nan_copy_pixel_to_argb32(struct Pixel pixel, char *dest)
{
	nan_copy_pixel_to_rgb24(pixel, dest);
	dest[3] = 255 * pixel.rgb[3];
}

void nan_img_copy_pixel_to_argb32(Image *img, int x, int y, char *dest)
{
	struct Pixel pix = nan_img_get_pixel(img, x, y);
	nan_copy_pixel_to_argb32(pix, dest);
}

void nan_copy_pixel_to_rgb24(struct Pixel pixel, char *dest)
{
	int i;
	for(i = 0; i < 3; ++i) {
		dest[2 - i] = 255 * pixel.rgb[i];
	}

}

void nan_img_copy_pixel_to_rgb24(Image *img, int x, int y, char *dest)
{
	struct Pixel pix = nan_img_get_pixel(img, x, y);
	nan_copy_pixel_to_rgb24(pix, dest);
}

void print_region(Rectangle region, const char * label)
{
	fprintf(stderr, "%s: (%d, %d) w: %d, h: %d)\n", label, region.x, region.y, region.w, region.h);
}

void nan_img_draw_circle(Image *img, int x, int y, int radius, int aaradius)
{
	int j;
	if(radius <= 0) {
		return ;
	}

	int mid = radius / 2;
	for(j = (int) (y - radius); j < (int) (y + radius); ++j) {
		int i;
		for(i = (int) (x - radius); i < (int) (x + radius); ++i) {
			double currad = distance(i - x, j - y);
			double raddiff = fabs(currad - radius);
			if(raddiff > aaradius) {
				continue;
			}

			double alpha = 1 - (raddiff / aaradius);
			nan_img_set_pixel(img, i, j, rgbpixel_with_alpha(PIXEL_BLACK, alpha));
		}

	}

}

Rectangle rectangle_default()
{
	Rectangle s;
	s.x = 0;
	s.y = 0;
	s.w = 0;
	s.h = 0;

	return s;
}

DrawingParams drawing_params_default()
{
	DrawingParams s;
	s.active_tool = TOOL_BRUSH;
	s.brushparams = brush_params_default();
	s.fgcolor = pixel_default();
	s.erase = false;
	s.pressure = 0.0;
	s.oldpressure = 0.0;

	return s;
}

