/* core.h
 * This file is part of Vara digital painting program
 * Copyright 2021, 2023 Nandakumar Edamana
 * Distributed under GNU General Public License, version 3
 */ 

#ifndef N96_DRAW_core_h
#define N96_DRAW_core_h

#include "warpcore.h"
#include "brush-params.h"
#ifdef _WIN32
static inline double drand48() { return rand() / (double) RAND_MAX; }
#endif
extern struct Pixel PIXEL_BLACK;
extern struct Pixel PIXEL_GRAY;
extern struct Pixel PIXEL_RED;
extern struct Pixel PIXEL_VPBG;
extern struct Pixel PIXEL_WHITE;
extern struct Pixel PIXEL_TRANSP;
extern struct Pixel PIXEL_CHECKER_DARK;
extern struct Pixel PIXEL_CHECKER_LIGHT;
#define AA_RADIUS_CIRCLE1 (2)
#define CIRCLE1_AA_IS_INSIDE (true)
extern double AA_RADIUS_CIRCLE2;
extern double AA_MAXDIFF_LINE;
#define NOISEDIVISOR (10)
typedef enum Tool {
	TOOL_BRUSH,
	TOOL_EYEDROPPER,
	TOOL_FLOODFILL,
	TOOL_PENCIL,
} Tool;

typedef struct Rectangle {
	int x;
	int y;
	int w;
	int h;
} Rectangle;

typedef struct DrawingParams {
	Tool active_tool;
	struct BrushParams brushparams;
	struct Pixel fgcolor;
	_Bool erase;
	double pressure;
	double oldpressure;
} DrawingParams;

typedef struct Image Image;
typedef struct _ngg_vtab_t_Image {
} _ngg_vtab_t_Image;

typedef struct Image {
	_ngg_vtab_t_Image _ngg_vtab_image;
	int width;
	int height;
	int depth;
	double *buf;
} Image;

Rectangle rectangle_new_from_boundaries(int x1, int y1, int x2, int y2);
_Bool region_is_zero(Rectangle region);
Rectangle region_new(int x, int y, int w, int h);
void image_construct(Image *this, int width, int height, int depth);
Image* image_Clone(Image *this);
void image_assert_region(Image *this, Rectangle region);
void image_copy_image(Image *this, Image *src);
Image* image_make_clone(Image *this);
Rectangle image_copy_image_region_with_destoffset(Image *this, Image *src, Rectangle region, int destx, int desty);
Rectangle image_copy_image_region_with_scale(Image *this, Image *src, Rectangle region, int destx, int desty, double scl);
Rectangle image_copy_image_with_destoffset(Image *this, Image *src, int destx, int desty);
Rectangle image_copy_image_with_scale(Image *this, Image *src, int destx, int desty, double scl);
Rectangle image_clip_region(Image *this, Rectangle region);
void image_copy_image_region(Image *this, Image *src, Rectangle region);
Rectangle image_dip(Image *this, double x, double y, DrawingParams dp);
void image_draw_debug_region(Image *this, Rectangle region);
Rectangle image_get_dirty_region_by_stroke(Image *this, double lbx, double lby, double ubx, double uby, DrawingParams dp);
Rectangle image_get_region(Image *this);
void image_fill(Image *this, struct Pixel pixel);
void image_fill_region(Image *this, struct Pixel pixel, Rectangle region);
_Bool image_is_point_outside(Image *this, int x, int y);
Rectangle image_line(Image *this, double x1, double y1, double x2, double y2, DrawingParams dp);
void image_overlay_image(Image *this, Image *img);
void image_overlay_image_region(Image *this, Image *img, Rectangle region);
void image_overlay_image_region_with_alpha(Image *this, Image *img, double alpha, Rectangle region);
void image_overlay_image_with_alpha(Image *this, Image *img, double alpha);
void image_overlay_image_with_offset(Image *this, Image *src, int destx, int desty);
void image_alphen_image_with_offset(Image *this, Image *src, int destx, int desty);
void image_destruct(Image *this);
double effective_brush_radius(double radius, DrawingParams dp);
Image* nan_img_new(int width, int height, int depth);
void nan_img_aaline_dot(Image *img, int x, int y, double diff, DrawingParams dp);
double distance(double xdist, double ydist);
double dist2pixmid(double x1, double y1, int x2, int y2);
int imax2(int a, int b);
int imin2(int a, int b);
void nan_img_aaline_inner_by_x_dbl(Image *img, double x1, double y1, double x2, double y2, double lbx, double lby, double ubx, double uby, double m, DrawingParams dp);
void nan_img_aaline_inner_by_y_dbl(Image *img, double x1, double y1, double x2, double y2, double lbx, double lby, double ubx, double uby, double m, DrawingParams dp);
void nan_img_thickline_inner_by_x(Image *img, double x1, double y1, double x2, double y2, double lbx, double lby, double ubx, double uby, double m, DrawingParams dp);
void nan_img_thickline_inner_by_y(Image *img, double x1, double y1, double x2, double y2, double lbx, double lby, double ubx, double uby, double m, DrawingParams dp);
void nan_img_thick_dot(Image *img, double x, double y, double progress, DrawingParams dp);
void nan_img_fill_circle1(Image *img, double x, double y, double brush_size, double opacity, _Bool noisy, DrawingParams dp);
void nan_img_fill_circle2(Image *img, double x, double y, double brush_size, double opacity, DrawingParams dp);
void nan_img_fill_random_square(Image *img, double x, double y, double brush_size, double opacity, DrawingParams dp);
_Bool pixelsmatch(struct Pixel px1, struct Pixel px2);
void nan_copy_pixel_to_argb32(struct Pixel pixel, char *dest);
void nan_img_copy_pixel_to_argb32(Image *img, int x, int y, char *dest);
void nan_copy_pixel_to_rgb24(struct Pixel pixel, char *dest);
void nan_img_copy_pixel_to_rgb24(Image *img, int x, int y, char *dest);
void print_region(Rectangle region, const char * label);
void nan_img_draw_circle(Image *img, int x, int y, int radius, int aaradius);
Rectangle rectangle_default();
DrawingParams drawing_params_default();

#endif /* N96_DRAW_core_h */
