/* quickpalette.c
 * This file is part of Vara digital painting program
 * Copyright 2021, 2023 Nandakumar Edamana
 * Distributed under GNU General Public License, version 3
 */ 

#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>

#include "quickpalette.h"

void quick_palette_construct(QuickPalette *this)
{
	image_view_construct((ImageView *) this);
	this->curh = 0;
	this->curw = 0;
	this->oncolorset_data = NULL;
	this->oncolorset_callback = NULL;
	this->autoresize = false;
	this->selcolor = PIXEL_RED;
	g_signal_connect((GtkWidget *) this->da, "size-allocate", G_CALLBACK(quick_palette_on_da_sizeallocate), this);
	gtk_widget_set_size_request((GtkWidget*) this->da, 240, 48);
	gtk_widget_add_events((GtkWidget *) this->da, GDK_BUTTON_PRESS_MASK);
	g_signal_connect((GtkWidget *) this->da, "event", G_CALLBACK(quick_palette_on_da_event), this);
}

_Bool quick_palette_on_da_event(GtkWidget *widget, GdkEvent *event, void *user_data)
{
	if(event->type != GDK_BUTTON_PRESS) {
		return;	}

	struct _ngg_tuple_ngg_coords_from_event _ngg_tmp_0 = ngg_coords_from_event(event);
	double y = _ngg_tmp_0.m1;
	double x = _ngg_tmp_0.m0;
	struct _ngg_tuple_get_color_at _ngg_tmp_1 = quick_palette_get_color_at(((QuickPalette *) user_data), x, y);
	_Bool in_spectrum = _ngg_tmp_1.m2;
	_Bool valid = _ngg_tmp_1.m1;
	struct Pixel color = _ngg_tmp_1.m0;
	if(!valid) {
		return;	}

	if(in_spectrum && (!(pixelsmatch(color, ((QuickPalette *) user_data)->selcolor)))) {
		fill_spectrum(((QuickPalette *) user_data)->img, color, false);
		image_view_queue_draw((ImageView *) ((QuickPalette *) user_data));
		((QuickPalette *) user_data)->selcolor = color;
	}

	if(((QuickPalette *) user_data)->oncolorset_callback) {
		assert(1 == color.rgb[3]); /* quickpalette.ngg:51 */
		((QuickPalette *) user_data)->oncolorset_callback(((QuickPalette *) user_data), color, ((QuickPalette *) user_data)->oncolorset_data);
	}

}

void quick_palette_on_da_sizeallocate(GtkDrawingArea *widget, GdkRectangle *nggevent_allocation, void *user_data)
{
	int neww = nggevent_allocation->width;
	int newh = nggevent_allocation->height;
	neww -= (neww % (imax2(6, NSHADES)));
	newh -= (newh % 4);
	if((neww <= 0) || (newh <= 0)) {
		return;	}

	if((neww == ((QuickPalette *) user_data)->curw) && (newh == ((QuickPalette *) user_data)->curh)) {
		return;	}


	Image * img2 = malloc(sizeof(Image));
	if(img2 == NULL) { perror(NULL); exit(EXIT_FAILURE); }
	image_construct(img2, neww, newh, 255);
	fill_spectrum(img2, ((QuickPalette *) user_data)->selcolor, true);
	image_view_set_image((ImageView *) ((QuickPalette *) user_data), img2);
	((QuickPalette *) user_data)->curw = neww;
	((QuickPalette *) user_data)->curh = newh;
}

void quick_palette_set_hexpand(QuickPalette *this, _Bool expand)
{
	gtk_widget_set_hexpand((GtkWidget *) this->da, expand);
}

_ngg_tuple_get_color_at quick_palette_get_color_at(QuickPalette *this, int x, int y)
{
	if((x >= this->img->width) || (y >= this->img->height)) {
		return (_ngg_tuple_get_color_at){PIXEL_BLACK, false, false};
	}

	int spectrumh = this->img->height / 4;

	return (_ngg_tuple_get_color_at){nan_img_get_pixel(this->img, x, y), true, y < spectrumh};
}

void quick_palette_destruct(QuickPalette *this)
{
	if(this->img) {
		image_destruct(this->img);
		free(this->img);
	}

	this->img = NULL;
	image_view_destruct((ImageView *) this);
}

void fill_spectrum(Image *img, struct Pixel basecolor, _Bool redraw_spectrum)
{
	int gradtype;
	int y;
	int gx;
	int rx;
	int bx;
	int segwidth = img->width / 6;
	int shadewidth = img->width / NSHADES;
	int spectrumh = img->height / 4;
	struct Rectangle region = image_get_region(img);
	if(!redraw_spectrum) {
		region.y = spectrumh;
		region.h -= spectrumh;
	}

	image_fill_region(img, PIXEL_WHITE, region);
	_Bool steppy = true;
	for(gradtype = 1; gradtype < 4; ++gradtype) {
		int x;
		struct Pixel _ngg_tmp_2;
		if(gradtype == 1) {
			_ngg_tmp_2 = PIXEL_BLACK;
		} else {
			struct Pixel _ngg_tmp_3;
			if(gradtype == 2) {
				_ngg_tmp_3 = PIXEL_GRAY;
			} else {
				_ngg_tmp_3 = PIXEL_WHITE;
			}
			_ngg_tmp_2 = _ngg_tmp_3;
		}
		struct Pixel fromclr = _ngg_tmp_2;
		struct Pixel curclr = fromclr;
		for(x = 0; x < img->width; ++x) {
			int y;
			double xfac = ((double) x) / img->width;
			if((!steppy) || (0 == (x % shadewidth))) {
				int i;
				for(i = 0; i < 3; ++i) {
					curclr.rgb[i] = fromclr.rgb[i] + (xfac * (basecolor.rgb[i] - fromclr.rgb[i]));
				}

			}

			int yoffset = spectrumh * gradtype;
			for(y = yoffset; y < yoffset + spectrumh; ++y) {
				nan_img_set_pixel(img, x, y, curclr);
			}

		}

	}

	for(y = spectrumh; y < img->height; ++y) {
		if(0 == (y % spectrumh)) {
			int x;
			for(x = 0; x < img->width; ++x) {
				nan_img_set_pixel(img, x, y, PIXEL_BLACK);
			}

		} else {
			int x;
			for(x = 0; x < img->width; x += shadewidth) {
				nan_img_set_pixel(img, x, y, PIXEL_BLACK);
			}

		}

	}

	if(!redraw_spectrum) {
		return ;
	}

	int steps = 4;
	for(gx = 0; gx <= segwidth; ++gx) {
		double g = ((double) ((gx >> steps) << steps)) / segwidth;
		for(y = 0; y < spectrumh; ++y) {
			nan_img_set_pixel(img, gx, y, nan_pixel_new(1, g, 0, 1));
		}

	}

	for(rx = 0; rx <= segwidth; ++rx) {
		double r = 1 - (((double) ((rx >> steps) << steps)) / segwidth);
		for(y = 0; y < spectrumh; ++y) {
			nan_img_set_pixel(img, segwidth + rx, y, nan_pixel_new(r, 1, 0, 1));
		}

	}

	for(bx = 0; bx <= segwidth; ++bx) {
		double b = ((double) ((bx >> steps) << steps)) / segwidth;
		for(y = 0; y < spectrumh; ++y) {
			nan_img_set_pixel(img, (2 * segwidth) + bx, y, nan_pixel_new(0, 1, b, 1));
		}

	}

	for(gx = 0; gx <= segwidth; ++gx) {
		double g = 1 - (((double) ((gx >> steps) << steps)) / segwidth);
		for(y = 0; y < spectrumh; ++y) {
			nan_img_set_pixel(img, (3 * segwidth) + gx, y, nan_pixel_new(0, g, 1, 1));
		}

	}

	for(rx = 0; rx <= segwidth; ++rx) {
		double r = ((double) ((rx >> steps) << steps)) / segwidth;
		for(y = 0; y < spectrumh; ++y) {
			nan_img_set_pixel(img, (4 * segwidth) + rx, y, nan_pixel_new(r, 0, 1, 1));
		}

	}

	for(bx = 0; bx <= segwidth; ++bx) {
		double b = 1 - (((double) ((bx >> steps) << steps)) / segwidth);
		for(y = 0; y < spectrumh; ++y) {
			nan_img_set_pixel(img, (5 * segwidth) + bx, y, nan_pixel_new(1, 0, b, 1));
		}

	}

}

